/*
 * This file is part of Hopsworks
 * Copyright (C) 2018, Logical Clocks AB. All rights reserved
 *
 * Hopsworks is free software: you can redistribute it and/or modify it under the terms of
 * the GNU Affero General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version.
 *
 * Hopsworks is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License along with this program.
 * If not, see <https://www.gnu.org/licenses/>.
 */

package io.hops.hopsworks.common.tensorflow;

import com.google.common.base.Strings;
import io.hops.hopsworks.common.dao.project.Project;
import io.hops.hopsworks.common.dao.python.CondaCommandFacade;
import io.hops.hopsworks.common.dao.python.CondaCommands;
import io.hops.hopsworks.common.dao.tensorflow.TfLibMapping;
import io.hops.hopsworks.common.dao.tensorflow.TfLibMappingFacade;
import io.hops.hopsworks.common.python.environment.EnvironmentController;
import io.hops.hopsworks.common.util.Settings;

import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import java.io.File;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Obs! This code is here as it should be used both for Jupyter and for running
 * notebooks as jobs
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.NEVER)
public class TfLibMappingUtil {

  private static final String LIB_PATH = "/usr/local";
  private static final String CUDA_BASE_PATH = LIB_PATH + "/cuda-";
  private static final String CUDNN_BASE_PATH = LIB_PATH + "/cudnn-";
  private static final String NCCL_BASE_PATH = LIB_PATH + "/nccl";
  @EJB
  private TfLibMappingFacade tfLibMappingFacade;
  @EJB
  private Settings settings;
  @EJB
  private EnvironmentController environmentController;

  private String buildTfLdLibraryPath(TfLibMapping tfLibMapping) {

    StringBuilder ldPathBuilder = new StringBuilder();

    if(!Strings.isNullOrEmpty(tfLibMapping.getCudnnVersion())) {
      ldPathBuilder.append(CUDNN_BASE_PATH + tfLibMapping.getCudnnVersion() + "/lib64" + File.pathSeparator);
    }

    if(!Strings.isNullOrEmpty(tfLibMapping.getCudaVersion())) {
      ldPathBuilder.append(CUDA_BASE_PATH + tfLibMapping.getCudaVersion() + "/lib64" + File.pathSeparator);
    }

    if(!Strings.isNullOrEmpty(tfLibMapping.getNcclVersion())) {
      ldPathBuilder.append(NCCL_BASE_PATH + tfLibMapping.getNcclVersion() + "/lib" + File.pathSeparator);
    }

    return ldPathBuilder.toString();
  }

  public String getTfLdLibraryPath(Project project) {
    // Get information about which version of TensorFlow the user is running
    TfLibMapping tfLibMapping = findTfMappingForProject(project);
    // No tensorflow-gpu
    if (tfLibMapping == null) {
      return "";
    }
    return buildTfLdLibraryPath(tfLibMapping);
  }

  public TfLibMapping findTfMappingForProject(Project project) {

    if (!project.getCondaEnv()) {
      return tfLibMappingFacade.findByTfVersion(settings.getTensorflowVersion());
    }

    CondaCommands command = environmentController.getOngoingEnvCreation(project);

    if(command == null) {
      return project.getPythonDepCollection().stream()
              .filter(dep -> dep.getDependency().equals("tensorflow") || dep.getDependency().equals("tensorflow-gpu"))
              .findAny()
              .map(tfDep -> tfLibMappingFacade.findByTfVersion(tfDep.getVersion()))
              .orElse(null);
    } else if(command.getOp().compareTo(CondaCommandFacade.CondaOp.CREATE) == 0) {
      return tfLibMappingFacade.findByTfVersion(settings.getTensorflowVersion());
    } else if(command.getOp().compareTo(CondaCommandFacade.CondaOp.YML) == 0) {
      String envYml = command.getEnvironmentYml();

      Pattern tfCPUPattern = Pattern.compile("(tensorflow==\\d*.\\d*.\\d*)");
      Matcher tfCPUMatcher = tfCPUPattern.matcher(envYml);

      if(tfCPUMatcher.find()) {
        String [] libVersionPair = tfCPUMatcher.group(0).split("==");
        return tfLibMappingFacade.findByTfVersion(libVersionPair[1]);
      }

      Pattern tfGPUPattern = Pattern.compile("(tensorflow-gpu==\\d*.\\d*.\\d*)");
      Matcher tfGPUMatcher = tfGPUPattern.matcher(envYml);

      if(tfGPUMatcher.find()) {
        String [] libVersionPair = tfGPUMatcher.group(0).split("==");
        return tfLibMappingFacade.findByTfVersion(libVersionPair[1]);
      }

      Pattern tfRocmPattern = Pattern.compile("(tensorflow-rocm==\\d*.\\d*.\\d*)");
      Matcher tfRocmMatcher = tfRocmPattern.matcher(envYml);

      if(tfRocmMatcher.find()) {
        String [] libVersionPair = tfGPUMatcher.group(0).split("==");
        return tfLibMappingFacade.findByTfVersion(libVersionPair[1]);
      }
    }
    return null;
  }
}
